from typing import List

# In order to be compatible with different langchain versions
try:
    from langchain.callbacks.manager import CallbackManager
except ImportError:
    from langchain.callbacks import CallbackManager
from langchain.chat_models import ChatOpenAI

from lcserve import serving

try:
    from .helper import CustomTool, PredefinedTools, get_agent, get_tools
except ImportError:
    from helper import CustomTool, PredefinedTools, get_agent, get_tools


@serving(websocket=True)
async def autogpt(
    name: str,
    role: str,
    goals: List[str],
    predefined_tools: PredefinedTools = None,
    custom_tools: List[CustomTool] = None,
    human_in_the_loop: bool = False,
    **kwargs,
) -> str:
    websocket = kwargs["websocket"]
    streaming_handler = kwargs['streaming_handler']

    llm = ChatOpenAI(
        temperature=0,
        verbose=True,
        streaming=True,
        callback_manager=CallbackManager([streaming_handler]),
    )
    tools = get_tools(llm, predefined_tools, custom_tools)
    agent = get_agent(
        name=name,
        role=role,
        websocket=websocket,
        tools=tools,
        llm=llm,
        human_in_the_loop=human_in_the_loop,
    )
    return await agent.arun(goals)



from enum import Enum

from langchain.memory import ChatMessageHistory

try:
    from .helper import grouper
except ImportError:
    from helper import grouper


class MemoryMode(str, Enum):
    SUMMARY = "summary"
    SUMMARY_BUFFER = "summary_buffer"
    LLAMA_SUMMARY = "llama_summary"


def get_memory(history: ChatMessageHistory, mode=MemoryMode.SUMMARY_BUFFER):
    from langchain.llms import OpenAI

    if mode == MemoryMode.SUMMARY:
        from langchain.memory import ConversationSummaryMemory

        memory = ConversationSummaryMemory.from_messages(
            llm=OpenAI(temperature=0, verbose=True),
            chat_memory=history,
            memory_key="chat_history",
        )

    elif mode == MemoryMode.SUMMARY_BUFFER:
        from langchain.memory import ConversationSummaryBufferMemory

        memory = ConversationSummaryBufferMemory(
            llm=OpenAI(temperature=0, verbose=True),
            max_token_limit=2000,
            memory_key="chat_history",
            return_messages=True,
        )

        for first, second in grouper(history.messages, 2):
            outputs = (
                {second.role: second.content}
                if second is not None
                else {first.role: first.content}
            )
            memory.save_context(
                inputs={first.role: first.content},
                outputs=outputs,
            )

    elif mode == MemoryMode.LLAMA_SUMMARY:
        from llama_index import ListIndex

        try:
            from .llama import GPTMultiUserChatMemory
        except ImportError:
            from llama import GPTMultiUserChatMemory

        memory = GPTMultiUserChatMemory(
            index=ListIndex([]),
            llm=OpenAI(temperature=0, verbose=True),
            chat_memory=history,
            memory_key="chat_history",
            return_messages=True,
        )

        for first, second in grouper(history.messages, 2):
            outputs = (
                {second.role: second.content}
                if second is not None
                else {first.role: first.content}
            )
            memory.save_context(
                inputs={first.role: first.content},
                outputs=outputs,
            )

    return memory



from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pandas import DataFrame

import hubble


JINAAI_PREFIX = 'jinaai://'


def _import_pandas():
    try:
        import pandas as pd
    except ImportError:
        raise ImportError('Please install pandas using `pip install pandas`')
    return pd


def upload_df(df: 'DataFrame', name: str, to_csv_kwargs={}) -> str:
    with NamedTemporaryFile(suffix='.csv') as f:
        df.to_csv(f.name, **to_csv_kwargs)
        r = hubble.Client().upload_artifact(f=f.name, is_public=True, name=name)
        r.raise_for_status()
        id = r.json().get('data', {}).get('_id')
        if not id:
            raise ValueError('No id found in response')
        return JINAAI_PREFIX + id


def _download_df_from_jinaai(id: str, read_csv_kwargs={}) -> 'DataFrame':
    pd = _import_pandas()

    if not id.startswith(JINAAI_PREFIX):
        raise ValueError(f'Invalid id: {id}')
    id = id[len(JINAAI_PREFIX) :]

    with NamedTemporaryFile(suffix='.csv') as f:
        hubble.Client().download_artifact(id=id, f=f.name)
        return pd.read_csv(f.name, **read_csv_kwargs)


def download_df(id: str, read_csv_kwargs={}) -> 'DataFrame':
    pd = _import_pandas()

    if id.startswith(JINAAI_PREFIX):
        return _download_df_from_jinaai(id, read_csv_kwargs=read_csv_kwargs)
    else:
        df = pd.read_csv(id, **read_csv_kwargs)
        df.columns = [col.strip('" ') for col in df.columns]
        return df



import os
from typing import Dict, List

from rich.prompt import Prompt
from pydantic import BaseModel


class PredefinedTools(BaseModel):
    names: List[str]
    params: Dict[str, str]


class CustomTool(BaseModel):
    name: str
    prompt: str
    description: str


class UserInput(BaseModel):
    host: str
    objective: str
    first_task: str
    predefined_tools: PredefinedTools
    custom_tools: List[CustomTool]
    endpoint: str = 'baby_agi'
    interactive: bool = True
    envs: Dict[str, str] = {}


def prompt_user() -> UserInput:
    # Get host
    host = Prompt.ask("🌎 Enter host (e.g. wss://babyagi-bcd407883c.wolf.jina.ai)")

    # Check if host is provided, else exit
    if not host:
        print(
            "Deploy your own babyagi using `lc-serve deploy babyagi` and then run this app again."
        )
        exit()

    # Get objective
    objective = Prompt.ask("🧠 Enter objective: (e.g. Solve world hunger)")

    # Check if objective is provided, else exit
    if not objective:
        print("Please provide an objective.")
        exit()

    # Get first task
    first_task = Prompt.ask(
        "1️⃣  Enter first task (Default first task is `Make a list of action items`)",
        default="Make a list of action items",
    )

    # Get the names of predefined tools from user
    predefined_tools_names = Prompt.ask(
        "🧰 Enter predefined tools names (comma separated) (type 'done' when finished)",
        default="wikipedia",
    )
    predefined_tools_names_list = predefined_tools_names.split(",")

    # Get the params for predefined tools from user
    predefined_tools_params = Prompt.ask(
        "🧰 Enter predefined tools params (comma separated) (type 'done' when finished)",
        default="",
    )
    predefined_tools_params = predefined_tools_params.split(",")
    if len(predefined_tools_params) > 1:
        predefined_tools_params_dict = dict(
            [param.split(":") for param in predefined_tools_params]
        )
    else:
        predefined_tools_params_dict = {}

    # Get custom tools
    custom_tools_selected = []
    while True:
        if len(custom_tools_selected) == 0:
            create_tool = Prompt.ask("🔧 Add a custom tool? (y/n)", default="y")
        else:
            create_tool = Prompt.ask("🔧 More custom tools to add? (y/n)", default="n")

        if create_tool.lower() != "y":
            break
        tool_name = Prompt.ask("   Enter tool name", default="TODO")
        tool_prompt = Prompt.ask(
            "   Enter tool prompt",
            default="You are a planner who is an expert at coming up with a todo list for a given objective. Come up with a todo list of just 2 items for this objective: {objective}",
        )
        tool_description = Prompt.ask(
            "   Enter tool description",
            default="useful for when you need to come up with todo lists. Input: an objective to create a todo list for. Output: a todo list for that objective. Please be very clear what the objective is!",
        )
        custom_tools_selected.append(
            {"name": tool_name, "prompt": tool_prompt, "description": tool_description}
        )

    # Get envs
    envs_selected = {}
    while True:
        if len(envs_selected) == 0:
            add_env = Prompt.ask("💻 Add an environment variable? (y/n)", default="y")
        else:
            add_env = Prompt.ask("💻 More environment variables? (y/n)", default="n")

        if add_env.lower() != "y":
            break
        env_name = Prompt.ask("   Enter environment variable name")
        env_value = os.environ.get(env_name)
        if not env_value:
            env_value = Prompt.ask("   Enter environment variable value")
        envs_selected[env_name] = env_value

    return UserInput(
        **{
            "host": host,
            "objective": objective,
            "first_task": first_task,
            "predefined_tools": {
                "names": predefined_tools_names_list,
                "params": predefined_tools_params_dict,
            },
            "custom_tools": custom_tools_selected,
            "envs": envs_selected,
        }
    )
